require "../engine/*"
require "../ui/*"
require "../frequency_loop"
require "../imgui_helper"

require "crsfml"
require "imgui"
require "imgui-sfml"

require "log"
Log.setup(:debug)

class Log
  def self.debug(str)
    Log.debug { str }
  end
end

# TODO move this to engine
class Projectable(N)
  property :gravity_body, :color, :size, :outline_color, :outline_size
  def initialize(
       @gravity_body : Gravity::Body(N),
       @color : Tuple(Int32, Int32, Int32),
       @size : Int32,
       @outline_color : Tuple(Int32, Int32, Int32) = {0, 0, 0},
       @outline_size : Int32 = 0,
     )
  end

  def position
    @gravity_body.position
  end

  def acceleration
    @gravity_body.acceleration
  end

  def speed
    @gravity_body.speed
  end

  def sf_shape
    shape = SF::CircleShape.new(@size)
    shape.fill_color = SF.color(@color[0], @color[1], @color[2])

    shape.outline_thickness = @outline_size
    shape.outline_color = SF.color(@outline_color[0], @outline_color[1], @outline_color[2])
    shape.position = sf_position(0.5)
    shape
  end

  def sf_position(coef : Float64 = 1.0)
    SF.vector2(@gravity_body.position[0] * coef, @gravity_body.position[1] * coef)
  end
end

class Game
  @window : SF::RenderWindow
  @frequency_nano_rate : Int32
  @delta_clock : SF::Clock

  def initialize(framerate = 60, width = 800, height = 600)
    @window = SF::RenderWindow.new(
      SF::VideoMode.new(width, height),
      "Sandbox",
    )
    @delta_clock = SF::Clock.new
    ImGui::SFML.init(@window)
    @window.framerate_limit = framerate

    @frequency_nano_rate = 1_000_000_000 // framerate
    @frequency = Time::Span.new(nanoseconds: @frequency_nano_rate)
    @ui_states = {
      acceleration_digit: {
        :x => 0.0,
        :y => 0.0,
      },
      acceleration_log_analogic_pressed: [false], # note: I should probably use pointers instead of this shit
      acceleration_log_analogic_save: [0.0],
    }

    @bodies = {
      vessel: Gravity::MovingBody(2).new(mass: 1.0, g: 0.1, position: Vector[10.0, 10.0]),
      star: Gravity::MovingBody(2).new(mass: 300000.0, g: 0.1, position: Vector[400.0, 400.0]),
      pla1: Gravity::MovingBody(2).new(mass: 15000.0, g: 0.1, position: Vector[200.0, 200.0]),
      pla2: Gravity::MovingBody(2).new(mass: 18000.0, g: 0.1, position: Vector[280.0, 320.0]),
    }
    @vessel = Projectable(2).new(
      gravity_body: @bodies[:vessel],
      color: { 255, 0, 250 },
      size: 3,
    )
    @star = Projectable(2).new(
      gravity_body: @bodies[:star],
      color: { 200, 200, 50 },
      size: 10,
    )
    @pla1 = Projectable(2).new(
      gravity_body: @bodies[:pla1],
      color: { 150, 50, 50 },
      size: 5,
    )
    @pla2 = Projectable(2).new(
      gravity_body: @bodies[:pla2],
      color: { 50, 100, 150 },
      size: 5,
    )
    @g = Gravity::Field(2).new([@star.gravity_body, @pla1.gravity_body, @pla2.gravity_body])
    @timer = Physics::Tick::Timer.new
    @ui_events_handler = UI::EventHandler.new
  end

  def handle_events
    while event = @window.poll_event
      ImGui::SFML.process_event(@window, event)

      case event
      when SF::Event::Closed
        @window.close
        exit
      when SF::Event::KeyPressed
        puts "KeyPressed #{event}"
      when SF::Event::MouseButtonEvent
        puts "MouseButtonEvent #{event}"
      end

      @ui_events_handler.handle(event)
    end
  end

  def execute_loop
    while @window.open? # restart the loop when if it ends
      FrequencyLoop.new(@frequency).loop do |i|
        @window.clear(SF::Color::Black)
        handle_events
        ImGui::SFML.update(@window, @delta_clock.restart)
        execute(i)
        ImGui::SFML.render(@window)
        @window.display
      end
    end
  end

  include ImGui::Helper

  # will yield the value
  def draw_slider(label : String, data : Float64, max : Float64 = 100.0, min : Float64 = -100.0, flags : ImGui::ImGuiSliderFlags = ImGui::ImGuiSliderFlags::NoRoundToFormat, &)
    ptr = pointerof(data)
    if ImGui.slider_scalar(
        label: label,
        p_data: ptr,
        p_min: min,
        p_max: max,
        flags: flags,
      )
      yield data
    end
  end

  # will yield the value
  def draw_input(label : String, data : Float64, max : Float64 = 100.0, min : Float64 = -100.0, flags : ImGui::ImGuiInputTextFlags = ImGui::ImGuiInputTextFlags::EnterReturnsTrue, &)
    ptr = pointerof(data)
    if ImGui.input_scalar(
        label: label,
        p_data: ptr,
        flags: flags,
      )
      data = max if data > max
      data = min if data < min
      yield data
    end
  end

  def accelerate_by_tick(movable_body : Gravity::MovingBody)
    gravity_field = @g.acceleration(movable_body.position)
    acceleration = movable_body.acceleration + gravity_field
    Physics.move_by_one_tick!(@timer.tick, movable_body.position, movable_body.speed, acceleration)
  end

  def draw_pilot
    draw_table(title: "Trusters", headers: {"Name", "X", "Y", "unit"}) do
      draw_table_line(
        "Analogic acceleration",
        -> {
          draw_slider(
            label: "###acceleration_analogue_x",
            data: @vessel.acceleration[0],
            flags: (
              ImGui::ImGuiSliderFlags::Logarithmic
            ),
          ) do |value|
            puts "x acceleration was #{@vessel.acceleration[0]}"
            @vessel.acceleration[0] = value
            puts "x acceleration is now #{@vessel.acceleration[0]}"
          end
        },
        -> {
          draw_slider(
            label: "###acceleration_analogue_y",
            data: @vessel.acceleration[1],
            flags: (
              ImGui::ImGuiSliderFlags::Logarithmic
            ),
          ) do |value|
            Log.debug "x acceleration was #{@vessel.acceleration[1]}"
            @vessel.acceleration[1] = value
            Log.debug "x acceleration is now #{@vessel.acceleration[1]}"
          end
        },
        "m/s²",
      )
      # automatically reset acceleration to 0, used for maneuvers
      draw_table_line(
        "Low thrust",
        -> {
          draw_slider(
            label: "###acceleration_low_analogic_x",
            data: @vessel.acceleration[0],
          ) do |value|
            if !@ui_states[:acceleration_log_analogic_pressed][0]
              @ui_states[:acceleration_log_analogic_save][0] = @vessel.acceleration[0]
            end
            @ui_states[:acceleration_log_analogic_pressed][0] = true
            @vessel.acceleration[0] = value
            @ui_events_handler.add(UI::EventHandler::Event::Type::MouseButtonReleased) do |callback, _|
              @ui_events_handler.remove(UI::EventHandler::Event::Type::MouseButtonReleased, callback)
              @ui_states[:acceleration_log_analogic_pressed][0] = false
              @vessel.acceleration[0] = @ui_states[:acceleration_log_analogic_save][0]
              nil
            end
          end
        },
        -> {
          draw_slider(
            label: "###acceleration_low_analogic_y",
            data: @vessel.acceleration[1],
          ) do |value|
            if !@ui_states[:acceleration_log_analogic_pressed][0]
              @ui_states[:acceleration_log_analogic_save][0] = @vessel.acceleration[1]
            end
            @ui_states[:acceleration_log_analogic_pressed][0] = true
            @vessel.acceleration[1] = value

            lock_callback = @ui_events_handler.add(UI::EventHandler::Event::Type::KeyPressed) do |callback, event|
              puts "Keypressed #{event}"
              @ui_states[:acceleration_log_analogic_save][0] = value
            end

            @ui_events_handler.add(UI::EventHandler::Event::Type::MouseButtonReleased) do |callback, _|
              @ui_events_handler.remove(UI::EventHandler::Event::Type::MouseButtonReleased, callback)
              @ui_events_handler.remove(UI::EventHandler::Event::Type::MouseButtonReleased, lock_callback)
              @ui_states[:acceleration_log_analogic_pressed][0] = false
              @vessel.acceleration[1] = @ui_states[:acceleration_log_analogic_save][0]
              nil
            end

          end
        },
        "m/s²",
      )
      draw_table_line(
        "Computer acceleration",
        -> {
          draw_input(
            label: "###acceleration_digit_x",
            data: @ui_states[:acceleration_digit][:x],
          ) do |value|
            @ui_states[:acceleration_digit][:x] = value
          end
        },
        -> {
          draw_input(
            label: "###acceleration_digit_y",
            data: @ui_states[:acceleration_digit][:y],
          ) do |value|
            @ui_states[:acceleration_digit][:y] = value
          end
        },
        "m/s²",
      )
      draw_table_line(
        "",
        -> {
          if ImGui.button("Comfirm")
            @vessel.acceleration[0] = @ui_states[:acceleration_digit][:x]
            @vessel.acceleration[1] = @ui_states[:acceleration_digit][:y]
          end
        },
        -> {
          if ImGui.button("Reset")
            @ui_states[:acceleration_digit][:x] = @vessel.acceleration[0]
            @ui_states[:acceleration_digit][:y] = @vessel.acceleration[1]
          end
        },
        -> {
          if ImGui.button("Zero")
            @ui_states[:acceleration_digit][:x] = 0.0
            @ui_states[:acceleration_digit][:y] = 0.0
          end
        },
      )
    end
  end

  def draw_navigation(gravity_field)
    # ImGui.text "Total Accele is  : [#{acceleration[0].round(4)}, #{acceleration[1].round(4)}] m/s²"
    ImGui.text "Pulse Accele is  : [#{@vessel.acceleration[0].round(4)}, #{@vessel.acceleration[1].round(4)}] m/s²"
    ImGui.text "GraviticBody is  : [#{gravity_field[0].round(4)}, #{gravity_field[1].round(4)}] m/s²"
    ImGui.text "Current speed is : #{@vessel.speed.magnitude.round(4)} m/s"

    draw_table(title: "Actual vectors", headers: {"Name", "X", "Y", "unit"}) do
      draw_table_line "Position", @vessel.position[0].round(3).to_s, @vessel.position[1].round(3).to_s, "m"
      draw_table_line "Speed", @vessel.speed[0].round(3).to_s, @vessel.speed[1].round(3).to_s, "m/s"
      draw_table_line "Acceleration", @vessel.acceleration[0].round(3).to_s, @vessel.acceleration[1].round(3).to_s, "m/s²"
    end
  end

  def execute(tick : Int)
    @timer.timeskip!(nanoseconds: @frequency_nano_rate)
    gravity_field = @g.acceleration(@bodies[:vessel].position)
    accelerate_by_tick(@bodies[:vessel])
    accelerate_by_tick(@bodies[:pla1])
    accelerate_by_tick(@bodies[:pla2])
    # accelerate_by_tick(@bodies[:star])

    if ImGui.begin("Bridge")
      if ImGui.tree_node_ex("Navigation", ImGui::ImGuiTreeNodeFlags.new(ImGui::ImGuiTreeNodeFlags::DefaultOpen))
        draw_navigation(gravity_field)
        ImGui.tree_pop
      end

      if ImGui.tree_node_ex("Pilot", ImGui::ImGuiTreeNodeFlags.new(ImGui::ImGuiTreeNodeFlags::DefaultOpen))
        draw_pilot
        ImGui.tree_pop
      end

      ImGui.end
    end

    @window.draw(@vessel.sf_shape)
    @window.draw(@star.sf_shape)
    @window.draw(@pla1.sf_shape)
    @window.draw(@pla2.sf_shape)
  end
end

Game.new(framerate: 60, width: 800, height: 600).execute_loop
